package hstar.creole.parser

import collection.mutable.{ListBuffer, MultiMap, HashMap, Set}
import hstar.creole.syntaxtree.{SimpleText, ForkElement, OList1, OList2, OList3, OList4, OList5, UOList5, UOList4, UOList3, UOList2, UOList1, AttributedElement, Element}

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 09.10.10<br>
 * Time: 19:32<br>
 */

trait FinalFoldingRule {
  def foldElements(elements:Seq[Element]):Seq[Element]
}

trait FoldingRule {
  def foldElements(elements: Seq[Element]): Seq[Element] = {
    if (elements.isEmpty) {
      Nil
    } else {
      //at first, group the elements into chunks that need to be folded and those that don't, but keep the order
      val chunked = {
        val chunkedElements = new ListBuffer[Chunk]
        var elementsToScan = elements.view
        do {
          if (chunkingRule(elementsToScan.head)) {
            chunkedElements += new Chunk(elementsToScan.takeWhile(chunkingRule), true)
          } else {
            chunkedElements += new Chunk(elementsToScan.takeWhile(e => !chunkingRule(e)), false)
          }
          elementsToScan = elementsToScan.drop(chunkedElements.last.elements.size)
        } while (elementsToScan.nonEmpty)
        chunkedElements.toList
      }
      val folded = chunked.flatMap(e => {
        if (e.matchedCondition) {
          foldChunkRule(e.elements)
        } else {
          e.elements
        }
      })
      folded
    }
  }

  protected val chunkingRule: (Element) => Boolean

  protected val foldChunkRule: (Seq[Element]) => Seq[Element]
}


private abstract class ListFoldingRule extends FoldingRule {
  private val foldOfSameType = new FoldElementsOfSameType
  
  protected val foldChunkRule = (elements: Seq[Element]) => {
    //this cast is safe because of our chunkingRule
    val listEntries = elements.asInstanceOf[Seq[AttributedElement]]

    def parentOf(element:AttributedElement) = {
      val ownDepth = syntax.depthOf(element.attribute)
      val parentIndex = listEntries.lastIndexWhere(e => syntax.depthOf(e.attribute) < ownDepth, listEntries.indexOf(element))
      if (parentIndex>=0) {
        Some(listEntries(parentIndex))
      } else {
        None
      }
    }
    // secret knowledge: after grouping the elements by their parent, the values still have the correct order because the inital list is an arraybuffer and therefore ordered
    val byParent = {
      //for each list entry, find its parent
      val elementToParent = listEntries.map(e => e -> parentOf(e))
      elementToParent.groupBy(_._2).map(e => e._1 -> e._2.map(e => e._1))
    }
    val asTree = {
      val sortedRootEntries = listEntries.filter(e => byParent.get(None).exists(rootElements => rootElements.contains(e)))
      sortedRootEntries.flatMap(parent => {
        def buildChildrenBranch(parent: AttributedElement): List[AttributedElement] = {
          //the foldOfSameType-rule will be applied to the result we return, but only to its first level - we need to take care of the lower levels ourselves
          val rawChildren = byParent.getOrElse(Some(parent), Nil)
          val childrenOfE = foldOfSameType.foldElements(rawChildren.flatMap(buildChildrenBranch))
          if (childrenOfE.size > 1) {
            val structuredChildren = ForkElement(childrenOfE)
            List(parent, AttributedElement(structuredChildren, parent.attribute))
          } else if (childrenOfE.isEmpty) {
            List(parent)
          } else {
            List(parent, AttributedElement(childrenOfE.head, parent.attribute))
          }
        }
        buildChildrenBranch(parent)
      })
    }
    asTree
  }

  protected def syntax: ListSyntax
}

private class UnorderedListFoldingRule extends ListFoldingRule {
  protected val chunkingRule = (e: Element) =>
    e match {
      case AttributedElement(_, UOList1) => true
      case AttributedElement(_, UOList2) => true
      case AttributedElement(_, UOList3) => true
      case AttributedElement(_, UOList4) => true
      case AttributedElement(_, UOList5) => true
      case _ => false
    }

  protected def syntax = SyntaxRules.unorderedListRule
}

private class OrderedListFoldingRule extends ListFoldingRule {
  protected val chunkingRule = (e: Element) =>
    e match {
      case AttributedElement(_, OList1) => true
      case AttributedElement(_, OList2) => true
      case AttributedElement(_, OList3) => true
      case AttributedElement(_, OList4) => true
      case AttributedElement(_, OList5) => true
      case _ => false
    }

  protected def syntax = SyntaxRules.orderedListRule

}

private class FoldElementsOfSameType extends FoldingRule {

  protected val foldChunkRule = (elements: Seq[Element]) => {
    if (elements.isEmpty) {
      Nil
    } else if (elements.size > 1) {
      val it = elements.iterator.asInstanceOf[Iterator[AttributedElement]]
      val result = new ListBuffer[Element]
      //split into all subsequences that share the same attribute and replace these with an attributedelement and an internal fork, but leave single elements as they are 
      while (it.hasNext) {
        val element = it.next()
        val followingElementsWithSameAttribute = it.takeWhile(_.attribute == element.attribute).toList
        val elementCount = followingElementsWithSameAttribute.size
        if (elementCount+1 == 1) {
          result += followingElementsWithSameAttribute.head
        } else if (elementCount+1 > 1) {
          val allWithSameAttribute = (element :: followingElementsWithSameAttribute).map(e => e.wrapped)
          result += AttributedElement(ForkElement(allWithSameAttribute), element.attribute)
          it.drop(elementCount)
        }
      }
      result
    } else {
      elements
    }
  }
  
  //we can handle anything that has attributes
  protected val chunkingRule = (e:Element) => e.isInstanceOf[AttributedElement]
}

private class FoldSimpleTexts extends FinalFoldingRule {


  def foldElements(elements: scala.Seq[Element]) = {
    List(SimpleText(("" /: elements)(_ + _.asInstanceOf[SimpleText].text)))
  }

  protected val chunkingRule = (e:Element) => e.isInstanceOf[SimpleText]
}

private class Chunk(val elements: Seq[Element], val matchedCondition: Boolean)


object FoldingRules {
  val allLiveRules: Traversable[FoldingRule] = new UnorderedListFoldingRule :: 
                                                  new OrderedListFoldingRule ::
                                                  new FoldElementsOfSameType ::
                                                  Nil

  val finalFolding: Traversable[FinalFoldingRule] = new FoldSimpleTexts ::
                                                    Nil
}