package hstar.creole.parser

import hstar.creole.syntaxtree._

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 02.10.2010<br>
 * Time: 19:49:04<br>
 */

trait SyntaxRule {

  protected lazy val canNotBeginInside: Seq[SyntaxRule] = Nil
  protected lazy val mustBeInsideElementToBegin         = Option.empty[Seq[SyntaxRule]]

  protected lazy val canNotContinueInside: Seq[SyntaxRule] = Nil
  protected lazy val mustBeInsideElementToContinue         = Option.empty[Seq[SyntaxRule]]

  override def toString = getClass.getSimpleName


  def canBegin(ctx: ParsingContext) = {
    val possibleByNestingRule = {
      def forbiddenBecauseOfParent = {
        canNotBeginInside.exists(e => isTagOfTypeOpen(ctx, e))
      }
      def explicitlyAllowedByParentOrNotRestricted = {
        mustBeInsideElementToBegin.map(_.forall(e => isTagOfTypeOpen(ctx, e))).getOrElse(true)
      }
      !forbiddenBecauseOfParent && explicitlyAllowedByParentOrNotRestricted
    }
    possibleByNestingRule && canStartAt(ctx)
  }

  def canContinue(ctx: ParsingContext) = {
    val possibleByNestingRule = {
      def forbiddenBecauseOfParent: Boolean = {
        canNotContinueInside.exists(e => isTagOfTypeOpen(ctx, e))
      }
      def explicitlyAllowedByParentOrNotRestricted: Boolean = {
        mustBeInsideElementToContinue.map(_.forall(e => isTagOfTypeOpen(ctx, e))).getOrElse(true)
      }
      !forbiddenBecauseOfParent && explicitlyAllowedByParentOrNotRestricted
    }
    possibleByNestingRule && canContinueAt(ctx)
  }

  protected[parser] def canStartAt(ctx: ParsingContext): Boolean

  protected def canContinueAt(ctx: ParsingContext): Boolean

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode): Element

  final def skipSyntaxInitMarker(cursor: Cursor, ctx: ParsingContext) = {
    val from = cursor.currentIndex
    consumeBeginMarker(cursor, ctx)
    ctx.source.substring(from, cursor.currentIndex)
  }

  final def skipSyntaxExitMarker(cursor: Cursor, ctx: ParsingContext) = {
    val from = cursor.currentIndex
    consumeEndMarker(cursor, ctx)
    ctx.source.substring(from, cursor.currentIndex)
  }

  protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = {}

  protected def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = {}

  def endsAllBranchesAt(ctx: ParsingContext) = {
    ctx.remainingText.isEmpty || endsAllBranchesAtNonFinalPosition(ctx)
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext): Boolean

  def foldParsedAndChildElements(ownElements: Seq[Element], children: Seq[Element]): Seq[Element] = {
    //default implementation, return the nonempty-sequence
    val returnThese = {
      if (ownElements.isEmpty || ownElements.forall(_.isInstanceOf[VoidElement])) {
        children
      } else if (children.isEmpty || children.forall(_.isInstanceOf[VoidElement])) {
        ownElements
      } else {
        throw new CreoleSyntaxError("default implementation cannot handle both children and own elements or none:" + children + ", own elements: " + ownElements)
      }
    }
    assert(returnThese.forall(!_.isInstanceOf[PlaceholderElement[_]]), "placeholder detected: " + returnThese)
    returnThese
  }

  protected final def isSameSyntaxAsCurrentElement(ctx: ParsingContext) = {
    val mostRecentNode = ctx.currentIncompleteElement.mostRecentlyAttachedNode(true)
    val isOpenOnSameLevel = mostRecentNode.exists(_.syntax == SyntaxRule.this)
    isOpenOnSameLevel
  }

  protected final def tagDepthOfType(ctx: ParsingContext, rule: SyntaxRule) = {
    ctx.allParents.count(e => e.syntax == rule)
  }

  protected final def isTagOfTypeOpen(ctx: ParsingContext, rule: SyntaxRule) = {
    val depth = tagDepthOfType(ctx, rule)
    assert(depth <= 1)
    depth == 1
  }

  protected final def isTagOpen(ctx: ParsingContext) = {
    isTagOfTypeOpen(ctx, SyntaxRule.this)
  }

  def canDirectlyNestElementOfType(syntax: SyntaxRule): Boolean

  def canSwitchToElementOfType(syntax: SyntaxRule): Boolean

}

private class SimpleAttributeRule(marker: String, attribute: Attribute) extends SyntaxRule {

  override protected lazy val canNotBeginInside = this :: Nil

  private val desc = "<text between elements of this type will be " + attribute + ">"

  protected def canContinueAt(ctx: ParsingContext) = !ctx.remainingText.startsWith(marker)

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    ctx.remainingText.startsWith(marker)
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    //the actual logic takes place in foldParsedAndChildElements
    new VoidElement(desc)
  }

  override def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) {
    cursor.incCursorIndexBy(marker.length)
  }

  override def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) {
    if (SyntaxRules.paragraphRule.canBegin(ctx)) {
      //nop, will be consumed by the paragraph-rule as a beginmarker
    } else if (ctx.remainingText.startsWith(marker)) {
      cursor.incCursorIndexBy(marker.length)
    } else {
      throw new CreoleSyntaxError("unexpected end marker for " + this + ": " + ctx.remainingText)
    }
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = {
    val shouldEndHereIfOpen = {
      val markerMatches = ctx.remainingText.startsWith(marker)
      val isAtParagraph = SyntaxRules.paragraphRule.canBegin(ctx)
      markerMatches || isAtParagraph
    }
    val isOpen = {
      isTagOpen(ctx) || isSameSyntaxAsCurrentElement(ctx)
    }
    shouldEndHereIfOpen && isOpen
  }

  override def foldParsedAndChildElements(ownElements: Seq[Element], children: Seq[Element]) = {
    assert(ownElements.size == 2 && ownElements.forall(_.isInstanceOf[VoidElement]), "whoops: " + ownElements)
    val folded = AttributedElement(ForkElement(children).asUnforkedIfOnlyOneElement, attribute)
    List(folded)
  }

  override def canSwitchToElementOfType(syntax: SyntaxRule) = false

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = syntax != this
}

private class NoSyntax extends SyntaxRule {
  protected def canContinueAt(ctx: ParsingContext) = {
    noOtherRuleApplies(ctx)
  }

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    noOtherRuleApplies(ctx)
  }

  private def noOtherRuleApplies(ctx: ParsingContext) = {
    !SyntaxRules.allRules.exists(_.canBegin(ctx))
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    SimpleText(parsedSourceOnCurrentLevel)
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = false

  override def canSwitchToElementOfType(syntax: SyntaxRule) = true

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = false

}

private class BoldRule extends SimpleAttributeRule("**", Bold)

private class ItalicRule extends SimpleAttributeRule("//", Italic)

private class RawRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = ctx.remainingText.startsWith("}}}") && isTagOpen(ctx)

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = AttributedElement(parsedSourceOnCurrentLevel, Raw)

  protected def canContinueAt(ctx: ParsingContext) = !ctx.remainingText.startsWith("}}}")

  protected[parser] def canStartAt(ctx: ParsingContext) = ctx.remainingText.startsWith("{{{")

  override def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = cursor.incCursorIndexBy(3)

  override def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = cursor.incCursorIndexBy(3)

  override def canDirectlyNestElementOfType(syntax: SyntaxRule) = false

  override def canSwitchToElementOfType(syntax: SyntaxRule) = false
}

private class LinkRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = ctx.remainingText.startsWith("}}}") && isTagOpen(ctx)

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    val split = parsedSourceOnCurrentLevel.split('|')
    if (split.length == 1) {
      LinkAndDescription(None, split(0))
    } else if (split.length == 2) {
      LinkAndDescription(Some(split(1)), split(0))
    } else {
      throw new CreoleSyntaxError("cannot understand \"" + parsedSourceOnCurrentLevel + "\"")
    }
  }

  protected def canContinueAt(ctx: ParsingContext) = !ctx.remainingText.startsWith("]]")

  protected[parser] def canStartAt(ctx: ParsingContext) = ctx.remainingText.startsWith("[[")

  override def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = cursor.incCursorIndexBy(2)

  override def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = cursor.incCursorIndexBy(2)

  override def canDirectlyNestElementOfType(syntax: SyntaxRule) = false

  override def canSwitchToElementOfType(syntax: SyntaxRule) = false
}

private class HeaderRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = ctx.remainingText.startsWith("=")

  override protected lazy val canNotBeginInside = SyntaxRules.tableRule :: Nil;

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    val headerType = ctx.getBeginMarker.length match {
      case 1 => Header1
      case 2 => Header2
      case 3 => Header3
      case 4 => Header4
      case 5 => Header5
      case 6 => Header6
    }
    AttributedElement(parsedSourceOnCurrentLevel, headerType)
  }

  protected def canContinueAt(ctx: ParsingContext) = !ctx.remainingText.startsWith("=")

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    ctx.remainingText.startsWith("=")
  }

  private def consumeAllEqualSigns(cursor: Cursor, ctx: ParsingContext) = {
    cursor.incCursorIndexBy(ctx.remainingText.view.takeWhile(_ == '=').size)
  }

  override def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = {
    consumeAllEqualSigns(cursor, ctx)
  }

  override def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = {
    consumeAllEqualSigns(cursor, ctx)
  }

  override def canDirectlyNestElementOfType(syntax: SyntaxRule) = false

  override def canSwitchToElementOfType(syntax: SyntaxRule) = false
}

abstract class ListSyntax(protected val beginMarkerChar: Char) extends SyntaxRule {
  private val beginMarker = beginMarkerChar.toString

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = {
    ctx.remainingText.startsWith(SyntaxRules.lineBreak)
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    val listType = listTypeByDepth(ctx.getBeginMarker.length)
    new PlaceholderElement("<text after this placeholder of this type will be a list entry of type <" + listType + ">", Some(listType))
  }

  def listTypeByDepth(depth: Int): Attribute

  def depthOf(attribute: Attribute) = {
    attribute match {
      case OList1 | UOList1 => 1
      case OList2 | UOList2 => 2
      case OList3 | UOList3 => 3
      case OList4 | UOList4 => 4
      case OList5 | UOList5 => 5
    }
  }

  protected def canContinueAt(ctx: ParsingContext) = {
    !ctx.remainingText.startsWith(SyntaxRules.lineBreak) && !canStartAt(ctx)
  }

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    ctx.remainingText.startsWith(beginMarker)
  }

  override protected def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = {
    cursor.incCursorIndex()
  }

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = {
    cursor.incCursorIndexBy(ctx.remainingText.view.takeWhile(_ == beginMarkerChar).size)
  }

  override def canDirectlyNestElementOfType(syntax: SyntaxRule) = syntax match {
    case e: ListSyntax => false
    case e: HeaderRule => false
    case _ => true
  }


  override def canSwitchToElementOfType(syntax: SyntaxRule) = false

  override def foldParsedAndChildElements(ownElements: Seq[Element], children: Seq[Element]) = {
    val folded = ownElements.zip(children).map(e => {
      // this is safe. we know, because we created the placeholder
      val attribute = e._1.asInstanceOf[PlaceholderElement[Attribute]].info.get
      AttributedElement(e._2, attribute)
    }).toList
    folded
  }
}

private class OrderedListRule extends ListSyntax('#') {
  def listTypeByDepth(depth: Int) = depth match {
    case 1 => OList1
    case 2 => OList2
    case 3 => OList3
    case 4 => OList4
    case 5 => OList5
  }
}

private class UnorderedListRule extends ListSyntax('*') {
  def listTypeByDepth(depth: Int) = depth match {
    case 1 => UOList1
    case 2 => UOList2
    case 3 => UOList3
    case 4 => UOList4
    case 5 => UOList5
  }


  override protected[parser] def canStartAt(ctx: ParsingContext) = {
    val followsOtherListEntry = ctx.currentIncompleteElement.mostRecentlyAttachedNode(true).exists(_.syntax == SyntaxRules.unorderedListRule)
    val cannotBeConfusedWithBoldEntry = ctx.remainingText.view.takeWhile(_ == beginMarkerChar).size != 2
    super.canStartAt(ctx) && (followsOtherListEntry || cannotBeConfusedWithBoldEntry)
  }
}

private class ForcedLineBreakRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = false

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = ForcedLineBreak()

  protected def canContinueAt(ctx: ParsingContext) = false

  protected[parser] def canStartAt(ctx: ParsingContext) = ctx.remainingText.startsWith("\\\\")

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) {
    cursor.incCursorIndexBy(2)
  }

  def canSwitchToElementOfType(syntax: SyntaxRule) = true

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = false
}

private class SimpleLineBreakRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = false

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = SimpleLineBreak()

  protected def canContinueAt(ctx: ParsingContext) = false


  override protected lazy val canNotBeginInside = SyntaxRules.orderedListRule :: SyntaxRules.unorderedListRule :: Nil

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    ctx.remainingText.startsWith(SyntaxRules.lineBreak)
  }

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) {
    cursor.incCursorIndex()
  }

  def canSwitchToElementOfType(syntax: SyntaxRule) = true

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = false
}

private class ParagraphRule extends SyntaxRule {
  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = false

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = Paragraph()

  protected def canContinueAt(ctx: ParsingContext) = false

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    val remainingText = ctx.remainingText
    val startsWithLineBreak = remainingText.startsWith(SyntaxRules.lineBreak)
    def followedByEmptyLine = {
      val nextLineBreak = remainingText.indexOf(SyntaxRules.lineBreak, 1)
      if (nextLineBreak != -1) {
        remainingText.substring(1, nextLineBreak).trim.isEmpty
      } else {
        false
      }
    }
    startsWithLineBreak && followedByEmptyLine
  }

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) {
    val skipCount = ctx.remainingText.view.takeWhile(Character.isWhitespace).size
    cursor.incCursorIndexBy(skipCount)
  }

  def canSwitchToElementOfType(syntax: SyntaxRule) = true

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = false
}

private class TableHeaderCellRule extends TableCellRule {
  override protected def wrapChildInTableElement(element: Element) = TableHeaderCell(element)

  override protected[parser] def canStartAt(ctx: ParsingContext) = ctx.remainingText.startsWith("|=")

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = {
    cursor.incCursorIndexBy(2)
  }
}

private class TableCellRule extends SyntaxRule with TableOps {

  protected def wrapChildInTableElement(element: Element): Element = TableCell(element)

  def canSwitchToElementOfType(syntax: SyntaxRule) = {
    // can only be followed by another cell
    syntax == this
  }

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = {
    true
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = {
    val pipeMatch = ctx.remainingText.startsWith("|")
    //if it's the final pipe of a row, let the row rule handle the branch ending
    def notFinalPipe = {
      !SyntaxRules.tableRowRule.endsAllBranchesAtNonFinalPosition(ctx) && 
      !SyntaxRules.tableRule.endsAllBranchesAtNonFinalPosition(ctx)
    }
    val result = pipeMatch && notFinalPipe
    result
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    //logic in fold
    new VoidElement("placeholder for cell content")
  }

  protected def canContinueAt(ctx: ParsingContext) = false

  // only contains children

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    val startMarkerMatched = ctx.remainingText.startsWith("|")
    def isEndMarker = {
      //check if there are only whitespaces
      val remainingToCheck = ctx.remainingLine.drop(1)
      remainingToCheck.forall(_.isWhitespace)
    }
    startMarkerMatched && !isEndMarker
  }

  override protected def consumeBeginMarker(cursor: Cursor, ctx: ParsingContext) = {
    cursor.incCursorIndex()
  }
}

private class TableRowRule extends SyntaxRule with TableOps {

  protected def wrapChildInTableElement(element: Element) = TableRow(element)

  def canSwitchToElementOfType(syntax: SyntaxRule) = {
    // can only be followed by another row
    syntax == this
  }

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = {
    syntax == SyntaxRules.tableHeaderCellRule || syntax == SyntaxRules.tableCellRule
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = {
    doesRowEndHere(ctx) && !SyntaxRules.tableRule.endsAllBranchesAtNonFinalPosition(ctx)
  }
  
  private[parser] def doesRowEndHere(ctx:ParsingContext) = {
    def rowEndMatch(hasPipe: Boolean) = {
      val remainingText = ctx.remainingText
      val whitespaceCount = {
        val operateOn = {
          if (hasPipe) {
            remainingText.drop(1)
          } else {
            remainingText
          }
        }
        operateOn.prefixLength(e => e != SyntaxRules.lineBreakChar && e.isWhitespace)
      }
      val indexToCheck = {
        val addToIndex = {
          if (hasPipe) {
            1
          } else {
            0
          }
        }
        whitespaceCount + addToIndex
      }
      if (remainingText.isDefinedAt(indexToCheck)) {
        //check if the char after the whitespaces is a linebreak
        remainingText.charAt(indexToCheck) == SyntaxRules.lineBreakChar
      } else {
        // end of string reached
        true
      }
    }
    val pipeMatch = ctx.remainingText.startsWith("|")
    val endMatch = rowEndMatch(pipeMatch)
    endMatch
  } 

  override protected def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = {
    //only consume the end marker if it's not the last row of the current table
    val lastRow = SyntaxRules.tableRule.endsAllBranchesAt(ctx)
    if (!lastRow) {
      consumeMarkerAtEndOfRow(cursor, ctx)
    }
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = new VoidElement("placeholder for table row")

  protected def canContinueAt(ctx: ParsingContext) = false

  // the row only contains children

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    (SyntaxRules.tableCellRule.canBegin(ctx) || SyntaxRules.tableHeaderCellRule.canBegin(ctx))
  }
}

private class TableRule extends SyntaxRule with TableOps {
  protected def wrapChildInTableElement(element: Element) = Table(element)

  def canSwitchToElementOfType(syntax: SyntaxRule) = {
    syntax != SyntaxRules.tableHeaderCellRule && syntax != SyntaxRules.tableCellRule && syntax != SyntaxRules.tableRowRule
  }

  def canDirectlyNestElementOfType(syntax: SyntaxRule) = {
    syntax == SyntaxRules.tableRowRule
  }

  override protected def consumeEndMarker(cursor: Cursor, ctx: ParsingContext) = {
    consumeMarkerAtEndOfRow(cursor, ctx)                
  }

  protected[parser] def endsAllBranchesAtNonFinalPosition(ctx: ParsingContext) = {
    val endOfRow = SyntaxRules.tableRowRule.asInstanceOf[TableRowRule].doesRowEndHere(ctx)
    def followedByAnotherRow = {
      ctx.nextLine.map(e => SyntaxRules.tableRowRule.canBegin(ctx.withNewSource(e))).getOrElse(false)
    }
    val result = endOfRow && !followedByAnotherRow
    result
  }

  def toElement(parsedSourceOnCurrentLevel: String, ctx: IntermediateSyntaxNode) = {
    //logic takes place in fold method
    new VoidElement("placeholder for table root")
  }

  protected def canContinueAt(ctx: ParsingContext) = {
    // the table itself is just the root element and does not contain anything except its children
    false
  }

  private def subElementCanStart(ctx: ParsingContext) = {
    SyntaxRules.tableRowRule.canBegin(ctx)
  }

  protected[parser] def canStartAt(ctx: ParsingContext) = {
    subElementCanStart(ctx)
  }
}

trait TableOps extends SyntaxRule {
  override def foldParsedAndChildElements(ownElements: scala.Seq[Element], children: scala.Seq[Element]) = {
    assert(ownElements.forall(_.isInstanceOf[VoidElement]))
    List(wrapChildInTableElement(ForkElement(children).asUnforkedIfOnlyOneElement))
  }

  protected def consumeMarkerAtEndOfRow(cursor: Cursor, ctx: ParsingContext) {
    if (ctx.remainingText.startsWith("|")) {
      cursor.incCursorIndex()
    }
    //consume all whitespaces
    val consumeCount = ctx.remainingText.prefixLength(_ != SyntaxRules.lineBreakChar)
    cursor.incCursorIndexBy(consumeCount)
    //consume the line break
    if (ctx.remainingText.startsWith(SyntaxRules.lineBreak)) {
      cursor.incCursorIndex()
    }
  }

  override protected lazy val canNotBeginInside = this :: Nil

  protected def wrapChildInTableElement(element: Element): Element
}


object SyntaxRules {
  val orderedListRule    : ListSyntax = new OrderedListRule
  val unorderedListRule  : ListSyntax = new UnorderedListRule
  val boldRule           : SyntaxRule = new BoldRule
  val italicRule         : SyntaxRule = new ItalicRule
  val rawRule            : SyntaxRule = new RawRule
  val headerRule         : SyntaxRule = new HeaderRule
  val linkRule           : SyntaxRule = new LinkRule
  val paragraphRule      : SyntaxRule = new ParagraphRule
  val breakRule          : SyntaxRule = new ForcedLineBreakRule
  val simpleLineBreakRule: SyntaxRule = new SimpleLineBreakRule
  val tableRule          : SyntaxRule = new TableRule
  val tableRowRule       : SyntaxRule = new TableRowRule
  val tableHeaderCellRule: SyntaxRule = new TableHeaderCellRule
  val tableCellRule      : SyntaxRule = new TableCellRule

  val allRules = {
    unorderedListRule ::
    boldRule ::
    italicRule ::
    rawRule ::
    headerRule ::
    orderedListRule ::
    linkRule ::
    paragraphRule ::
    breakRule ::
    simpleLineBreakRule ::
    tableRule ::
    tableRowRule ::
    tableHeaderCellRule ::
    tableCellRule ::
    Nil
  }

  val defaultSyntax: SyntaxRule = new NoSyntax

  val allRulesAndDefault = allRules :+ defaultSyntax

  val lineBreak = "\n"

  val lineBreakChar = '\n'

}