package hstar.creole.parser.test

import org.scalatest.FunSuite
import org.scalatest.matchers.ShouldMatchers
import hstar.creole.parser.Parser
import hstar.creole.syntaxtree._

/**
 * Developed with pleasure :)<br>
 * User: HoD<br>
 * Date: 02.10.2010<br>
 * Time: 20:35:46<br>
 */

class SimpleTests extends FunSuite with ShouldMatchers {

  test("2 tables divided") {
    val element = Parser("|first|second|\n\n|third|fourth|")
    val firstTable = Table(TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second"))))))
    val secondTable = Table(TableRow(ForkElement(List(TableCell(SimpleText("third")), TableCell(SimpleText("fourth"))))))
    val expected = ForkElement(List(firstTable, SimpleLineBreak(), secondTable))
    element should equal(expected)
  }
  
  test("2x2 table cells without ending pipes") {
    val element = Parser("|first|second\n|third|fourth")
    val expected = {
      val firstRow = TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second")))))
      val secondRow = TableRow(ForkElement(List(TableCell(SimpleText("third")), TableCell(SimpleText("fourth")))))
      Table(ForkElement(List(firstRow, secondRow)))
    }
    element should equal (expected)
  }
  test("2x2 table cells") {
    val element = Parser("|first|second|\n|third|fourth|")
    val expected = {
      val firstRow = TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second")))))
      val secondRow = TableRow(ForkElement(List(TableCell(SimpleText("third")), TableCell(SimpleText("fourth")))))
      Table(ForkElement(List(firstRow, secondRow)))
    }
    element should equal(expected)
  }
  
  test("2 table cells") {
    val element = Parser("|first|second|")
    val expected = Table(TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second"))))))
    element should equal(expected)
  }
  
  test("2 table cells without ending pipe") {
    val element = Parser("|first|second")
    val expected = Table(TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second"))))))
    element should equal(expected)
  }
  
  test("table cell without ending pipe") {
    val element = Parser("|first")
    val expected = Table(TableRow(TableCell(SimpleText("first"))))
    element should equal(expected)
  }
  
  test("3 table cells") {
    val element = Parser("|first|second|third|")
    val expected = Table(TableRow(ForkElement(List(TableCell(SimpleText("first")), TableCell(SimpleText("second")), TableCell(SimpleText("third"))))))
    element should equal(expected)
  }
  
  test("single table cell, equal sign header is not allowed in tables") {
    val element = Parser("| === hello world ===|")
    val expected = Table(TableRow(TableCell(SimpleText(" === hello world ==="))))
    element should equal(expected)
  }

  test("single table cell (header) ") {
    val element = Parser("|=hello world|")
    val expected = Table(TableRow(TableHeaderCell(SimpleText("hello world"))))
    element should equal(expected)
  }
  
  test("bold text") {
    val result = Parser("**hello world**")
    val expected = AttributedElement("hello world", Bold)

    result should equal(expected)
  }

  test("single table cell") {
    val element = Parser("|hello world|")
    val expected = Table(TableRow(TableCell(SimpleText("hello world"))))
    element should equal(expected)
  }

  test("italic should stop at paragraph") {
    val element = Parser("//hello\n\nnormal//world")
    val expected = ForkElement(List(AttributedElement(SimpleText("hello"), Italic), Paragraph(), SimpleText("normal"), AttributedElement(SimpleText("world"), Italic)))
    element should equal(expected)
  }

  test("bold should stop at paragraph") {
    val element = Parser("**hello\n\nnormal**world")
    val expected = ForkElement(List(AttributedElement(SimpleText("hello"), Bold), Paragraph(), SimpleText("normal"), AttributedElement(SimpleText("world"), Bold)))
    element should equal(expected)
  }

  test("italic text twice without whitespace") {
    val result = Parser("//hello////world//")
    val expected = AttributedElement(ForkElement(List(SimpleText("hello"),SimpleText("world"))), Italic)

    result should equal(expected)
  }

  test("read simple test") {
    val result = Parser("hello world")
    val expected = SimpleText("hello world")

    result should equal(expected)
  }



  test("bold text twice") {
    val result = Parser("**hello** **world**")
    val expected = ForkElement(List(AttributedElement(SimpleText("hello"), Bold), SimpleText(" "), AttributedElement(SimpleText("world"), Bold)))

    result should equal(expected)
  }

  test("italic text twice") {
    val result = Parser("//hello// //world//")
    val expected = ForkElement(List(AttributedElement(SimpleText("hello"), Italic), SimpleText(" "), AttributedElement(SimpleText("world"), Italic)))

    result should equal(expected)
  }


  test("bold text in simple text") {
    val result = Parser("hello**bold**world")
    val expected = ForkElement(List(SimpleText("hello"), AttributedElement("bold", Bold), SimpleText("world")))

    result should equal(expected)
  }

  test("bold text then italic text") {
    val result = Parser("hello**bold**world//italic//")
    val expected = ForkElement(List(SimpleText("hello"), AttributedElement("bold", Bold), SimpleText("world"), AttributedElement("italic", Italic)))

    result should equal(expected)
  }

  test("italic text") {
    val result = Parser("//hello world//")
    val expected = AttributedElement("hello world", Italic)

    result should equal(expected)
  }

  test("italic text in simple text") {
    val result = Parser("hello//italic//world")
    val expected = ForkElement(List(SimpleText("hello"), AttributedElement("italic", Italic), SimpleText("world")))

    result should equal(expected)
  }

  test("italic text wrapped around bold text") {
    val result = Parser("//hello**bold**world//")
    val expected = AttributedElement(ForkElement(List(SimpleText("hello"), AttributedElement("bold", Bold), SimpleText("world"))),
      Italic)

    result should equal(expected)
  }

  test("bold text wrapped around italic text") {
    val result = Parser("**hello//italic//world**")
    val expected = AttributedElement(ForkElement(List(SimpleText("hello"), AttributedElement("italic", Italic), SimpleText("world"))),
      Bold)

    result should equal(expected)
  }

  test("raw text") {
    val result = Parser("{{{**//check//**}}}")
    val expected = AttributedElement(SimpleText("**//check//**"), Raw)

    result should equal(expected)
  }

  test("header 1") {
    val result = Parser("=text=") :: Nil
    val expected = AttributedElement("text", Header1)
    checkAllEqualTo(result, expected)
  }

  test("header 2") {
    val result = Parser("==text=") :: Parser("==text==") :: Nil
    val expected = AttributedElement("text", Header2)
    checkAllEqualTo(result, expected)
  }

  test("header 3") {
    val result = Parser("===text=") :: Parser("===text==") :: Parser("===text===") :: Nil
    val expected = AttributedElement("text", Header3)
    checkAllEqualTo(result, expected)
  }

  test("header 4") {
    val result = Parser("====text=") :: Parser("====text==") :: Parser("====text===") :: Parser("====text====") :: Nil
    val expected = AttributedElement("text", Header4)
    checkAllEqualTo(result, expected)
  }

  test("header 5") {
    val result = Parser("=====text=") :: Parser("=====text==") :: Parser("=====text===") :: Parser("=====text====") :: Parser("=====text=====") :: Nil
    val expected = AttributedElement("text", Header5)
    checkAllEqualTo(result, expected)
  }

  test("header 6") {
    val result = Parser("======text=") :: Parser("======text==") :: Parser("======text====") :: Parser("======text=====") :: Parser("======text======") :: Nil
    val expected = AttributedElement("text", Header6)
    checkAllEqualTo(result, expected)
  }

  private def checkAllEqualTo(check: Traversable[Element], expected: Element) {
    check.foreach(e => e should equal(expected))
  }

  test("simple list entry") {
    val src = "#first level"
    val element = Parser(src)
    element should equal(AttributedElement(SimpleText("first level"), OList1))
  }


  test("two list entries") {
    val src = "#first entry\n#second entry"
    val element = Parser(src)
    element should equal(AttributedElement(ForkElement(List(SimpleText("first entry"), SimpleText("second entry"))), OList1))
  }

  test("two list entries with different depths") {
    val src = "#first entry\n##second entry"
    val element = Parser(src)
    element should equal(AttributedElement(ForkElement(List(SimpleText("first entry"), AttributedElement(SimpleText("second entry"), OList2))), OList1))
  }

  test("check if level 2 unordered list entry gets confused with bold text") {
    val src = "*first entry\n**second entry"
    val element = Parser(src)
    element should equal(AttributedElement(ForkElement(List(SimpleText("first entry"), AttributedElement(SimpleText("second entry"), UOList2))), UOList1))
  }

  test("simple ordered list") {
    val src = "#begin\n#level1\n##level2\n##level2_2\n#level1_2\n#end"
    val element = Parser(src)
    val expected = AttributedElement(ForkElement(List(SimpleText("begin"),
      SimpleText("level1"),
      AttributedElement(ForkElement(List(SimpleText("level2"), SimpleText("level2_2"))), OList2),
      SimpleText("level1_2"),
      SimpleText("end"))), OList1)

    element should equal(expected)
  }

  test("simple ordered list with many entries at level 2") {
    val src = "#begin\n#level1\n##level2\n##level2_2\n##level2_3\n##level2_4\n##level2_5\n##level2_6\n#level1_2\n#end"
    val element = Parser(src)
    val expected = AttributedElement(ForkElement(List(SimpleText("begin"),
      SimpleText("level1"),
      AttributedElement(ForkElement(List(SimpleText("level2"),
        SimpleText("level2_2"),
        SimpleText("level2_3"),
        SimpleText("level2_4"),
        SimpleText("level2_5"),
        SimpleText("level2_6"))), OList2),
      SimpleText("level1_2"),
      SimpleText("end"))), OList1)
    element should equal(expected)
  }

  test("all list levels down (ordered)") {
    val src = "#level1\n" +
              "##level2\n##level2_2\n" +
              "###level3\n###level3_2\n###level3_3\n" +
              "####level4\n####level4_2\n####level4_3\n####level4_4\n" +
              "#####level5\n#####level5_2\n#####level5_3\n#####level5_4\n#####level5_5"
    val element = Parser(src)

    val level5 = AttributedElement(ForkElement(List(SimpleText("level5"), SimpleText("level5_2"), SimpleText("level5_3"), SimpleText("level5_4"), SimpleText("level5_5"))),
      OList5)
    val level4 = AttributedElement(ForkElement(List(SimpleText("level4"), SimpleText("level4_2"), SimpleText("level4_3"), SimpleText("level4_4"), level5)), OList4)
    val level3 = AttributedElement(ForkElement(List(SimpleText("level3"), SimpleText("level3_2"), SimpleText("level3_3"), level4)), OList3)
    val level2 = AttributedElement(ForkElement(List(SimpleText("level2"), SimpleText("level2_2"), level3)), OList2)
    val expected = AttributedElement(ForkElement(List(SimpleText("level1"), level2)), OList1)
    element should equal(expected)
  }

  test("all list levels down (unordered)") {
    val src = "*level1\n" +
              "**level2\n**level2_2\n" +
              "***level3\n***level3_2\n***level3_3\n" +
              "****level4\n****level4_2\n****level4_3\n****level4_4\n" +
              "*****level5\n*****level5_2\n*****level5_3\n*****level5_4\n*****level5_5"
    val element = Parser(src)

    val level5 = AttributedElement(ForkElement(List(SimpleText("level5"), SimpleText("level5_2"), SimpleText("level5_3"), SimpleText("level5_4"), SimpleText("level5_5"))),
      UOList5)
    val level4 = AttributedElement(ForkElement(List(SimpleText("level4"), SimpleText("level4_2"), SimpleText("level4_3"), SimpleText("level4_4"), level5)), UOList4)
    val level3 = AttributedElement(ForkElement(List(SimpleText("level3"), SimpleText("level3_2"), SimpleText("level3_3"), level4)), UOList3)
    val level2 = AttributedElement(ForkElement(List(SimpleText("level2"), SimpleText("level2_2"), level3)), UOList2)
    val expected = AttributedElement(ForkElement(List(SimpleText("level1"), level2)), UOList1)
    element should equal(expected)
  }


  test("all list levels down and up again (ordered)") {
    val src = "#level1\n" +
              "##level2\n##level2_2\n" +
              "###level3\n###level3_2\n###level3_3\n" +
              "####level4\n####level4_2\n####level4_3\n####level4_4\n" +
              "#####level5\n#####level5_2\n#####level5_3\n#####level5_4\n#####level5_5\n" +
              "####level4x\n####level4_2x\n####level4_3x\n####level4_4x\n" +
              "###level3x\n###level3_2x\n###level3_3x\n" +
              "##level2x\n##level2_2x\n" +
              "#level1x"
    val element = Parser(src)
    val level5 = AttributedElement(ForkElement(List(SimpleText("level5"), SimpleText("level5_2"), SimpleText("level5_3"), SimpleText("level5_4"), SimpleText("level5_5"))),
      OList5)
    val level4 = AttributedElement(ForkElement(List(SimpleText("level4"),
      SimpleText("level4_2"),
      SimpleText("level4_3"),
      SimpleText("level4_4"),
      level5,
      SimpleText("level4x"),
      SimpleText("level4_2x"),
      SimpleText("level4_3x"),
      SimpleText("level4_4x"))), OList4)
    val level3 = AttributedElement(ForkElement(List(SimpleText("level3"),
      SimpleText("level3_2"),
      SimpleText("level3_3"),
      level4,
      SimpleText("level3x"),
      SimpleText("level3_2x"),
      SimpleText("level3_3x"))), OList3)
    val level2 = AttributedElement(ForkElement(List(SimpleText("level2"), SimpleText("level2_2"), level3, SimpleText("level2x"), SimpleText("level2_2x"))), OList2)
    val expected = AttributedElement(ForkElement(List(SimpleText("level1"), level2, SimpleText("level1x"))), OList1)
    element should equal(expected)
  }

  test("all list levels down and up again (unordered)") {
    val src = "*level1\n" +
              "**level2\n**level2_2\n" +
              "***level3\n***level3_2\n***level3_3\n" +
              "****level4\n****level4_2\n****level4_3\n****level4_4\n" +
              "*****level5\n*****level5_2\n*****level5_3\n*****level5_4\n*****level5_5\n" +
              "****level4x\n****level4_2x\n****level4_3x\n****level4_4x\n" +
              "***level3x\n***level3_2x\n***level3_3x\n" +
              "**level2x\n**level2_2x\n" +
              "*level1x"
    val element = Parser(src)
    val level5 = AttributedElement(ForkElement(List(SimpleText("level5"), SimpleText("level5_2"), SimpleText("level5_3"), SimpleText("level5_4"), SimpleText("level5_5"))),
      UOList5)
    val level4 = AttributedElement(ForkElement(List(SimpleText("level4"),
      SimpleText("level4_2"),
      SimpleText("level4_3"),
      SimpleText("level4_4"),
      level5,
      SimpleText("level4x"),
      SimpleText("level4_2x"),
      SimpleText("level4_3x"),
      SimpleText("level4_4x"))), UOList4)
    val level3 = AttributedElement(ForkElement(List(SimpleText("level3"),
      SimpleText("level3_2"),
      SimpleText("level3_3"),
      level4,
      SimpleText("level3x"),
      SimpleText("level3_2x"),
      SimpleText("level3_3x"))), UOList3)
    val level2 = AttributedElement(ForkElement(List(SimpleText("level2"), SimpleText("level2_2"), level3, SimpleText("level2x"), SimpleText("level2_2x"))), UOList2)
    val expected = AttributedElement(ForkElement(List(SimpleText("level1"), level2, SimpleText("level1x"))), UOList1)
    element should equal(expected)
  }

  test("simple link") {
    val src = "[[http://www.google.de]]"
    val element = Parser(src)
    element should equal(new LinkAndDescription(None, "http://www.google.de"))
  }

  test("named link") {
    val src = "[[http://www.google.de|The oracle]]"
    val element = Parser(src)
    element should equal(new LinkAndDescription(Some("The oracle"), "http://www.google.de"))
  }

  test("link description should not be markupped") {
    val src = "[[http://www.google.de|==//The **o**racle{{{==]]"
    val element = Parser(src)
    element should equal(new LinkAndDescription(Some("==//The **o**racle{{{=="), "http://www.google.de"))
  }


  test("forced line break") {
    val element = Parser("hello\\\\world")
    val expected = ForkElement(List(SimpleText("hello"), ForcedLineBreak(), SimpleText("world")))
    element should equal(expected)
  }

  test("simple line break") {
    val element = Parser("hello\nworld")
    val expected = ForkElement(List(SimpleText("hello"), SimpleLineBreak(), SimpleText("world")))
    element should equal(expected)
  }

  test("bold should not stop at line break") {
    val element = Parser("**hello\nworld**")
    val expected = AttributedElement(ForkElement(List(SimpleText("hello"), SimpleLineBreak(), SimpleText("world"))), Bold)
    element should equal(expected)
  }

  test("italic should not stop at line break") {
    val element = Parser("//hello\nworld//")
    val expected = AttributedElement(ForkElement(List(SimpleText("hello"), SimpleLineBreak(), SimpleText("world"))), Italic)
    element should equal(expected)
  }


}